# Copyright (c) Alibaba, Inc. and its affiliates.

import json
import pickle

import datasets
import numpy as np
from datasets import DownloadManager, Features, Value

from .base import CustomDatasetBuilder


class MCCEDatasetBuilderConfig(datasets.BuilderConfig):
    """BuilderConfig for entity typing datasets"""

    def __init__(self, data_dir=None, data_files=None, **corpus_config):
        super().__init__(data_dir=data_dir, data_files=data_files)
        self.corpus_config = corpus_config


class MCCEDatasetBuilder(CustomDatasetBuilder):
    """Builder for entity typing datasets.

    features:
        id: string, data record id.
        tokens: list[str] input tokens.
        spans: List[Dict],  mentions like:
        [{'start': 0, 'end': 2, 'type': ['PER', 'MAN'], 'candidates': ['PER', 'MAN', 'ORG']}]
        mask: bool, mention mask.
    """

    BUILDER_CONFIG_CLASS = MCCEDatasetBuilderConfig

    def stub():  # noqa: D102
        pass

    def _info(self):
        info = datasets.DatasetInfo(
            features=Features(
                {
                    'id': Value('string'),
                    'tokens': [Value('string')],
                    'spans': [
                        {
                            'start': Value('int32'),  # close
                            'end': Value('int32'),  # open
                            'type': [Value('string')],
                            'candidates': [Value('string')],
                        }
                    ],
                    'mask': [Value('bool')],
                }
            )
        )
        return info

    def _split_generators(self, dl_manager: DownloadManager):
        """Specify feature dictionary generators and dataset splits.

        This function returns a list of `SplitGenerator`s defining how to generate
        data and what splits to use.

        Example:
            the gen_kwargs should contain the typing dataset filename and the corresponding candidates
            return[
                    datasets.SplitGenerator(
                            name=datasets.Split.TRAIN,
                            gen_kwargs={'file': 'train_data.zip', 'cand': List[List[str]]},
                    ),
                    datasets.SplitGenerator(
                            name=datasets.Split.TEST,
                            gen_kwargs={'file': 'test_data.zip', 'cand': List[List[str]]},
                    ),
            ]

        The above code will first call `_generate_examples(file='train_data.zip', cand=...)`
        to write the train data, then `_generate_examples(file='test_data.zip', cand=...)` to
        write the test data.

        Datasets are typically split into different subsets to be used at various
        stages of training and evaluation.

        Note that for datasets without a `VALIDATION` split, you can use a
        fraction of the `TRAIN` data for evaluation as you iterate on your model
        so as not to overfit to the `TEST` data.

        For downloads and extractions, use the given `download_manager`.
        Note that the `DownloadManager` caches downloads, so it is fine to have each
        generator attempt to download the source data.

        A good practice is to download all data in this function, and then
        distribute the relevant parts to each split with the `gen_kwargs` argument

        Args:
            dl_manager: (DownloadManager) Download manager to download the data
            data should additionally contain a candidate file:
            {'train': List[List[str]], 'dev': List[List[str]], 'test': List[List[str]]},
            in which containing the candidates of each sample in the training/dev/test set.

        Returns:
            `list<SplitGenerator>`.
        """

        data_files = self._resolve_datasets(dl_manager=dl_manager)

        if 'cand' not in data_files:
            return [
                datasets.SplitGenerator(
                    name=split_name, gen_kwargs={'filepath': data_files[split_name], 'cand': None}
                )
                for split_name in data_files.keys()
            ]
        else:
            candidates = data_files.pop('cand')
            candidates = pickle.load(open(candidates, 'rb'))
            keys = candidates.keys()
            split_names = data_files.keys()
            for i in split_names:
                if i == 'valid' and i not in keys:
                    assert 'dev' in keys
                    candidates[i] = candidates['dev']
                elif i == 'dev' and i not in keys:
                    assert 'valid' in keys
                    candidates[i] = candidates['valid']
                else:
                    assert i in keys

            return [
                datasets.SplitGenerator(
                    name=split_name,
                    gen_kwargs={'filepath': data_files[split_name], 'cand': candidates[split_name]},
                )
                for split_name in data_files.keys()
            ]

    def _generate_examples(self, filepath, cand):
        """Load data file.
        Args:
        file_path: string, data file path.
        corpus_config: dictionary, required keys:
        tokenizer: string, specify the tokenization method,
            'char' list(text) and 'blank' for text.split(' ')
        is_end_included: bool, wheather the end index pointer to the
            last token or the token next to the last token.
            e.g., text = 'This is an example.', mention = 'example',
            end = 3 if is_end_included is True.

        """
        corpus_config = self.config.corpus_config
        spans_key = corpus_config.get('spans_key', 'label')
        text_key = corpus_config.get('text_key', 'text')
        tokenizer = corpus_config.get('tokenizer', 'char')
        is_end_included = corpus_config.get('is_end_included', False)

        with open(filepath, encoding='utf-8') as f:
            guid = 0
            for line in f:
                example = json.loads(line)
                text = example[text_key]
                if isinstance(text, list):
                    tokens = text
                elif isinstance(text, str):
                    if tokenizer == 'char':
                        tokens = list(text)
                    elif tokenizer == 'blank':
                        tokens = text.split()
                    else:
                        raise NotImplementedError
                else:
                    raise ValueError('Unsupported text input.')

                entities = list()
                for span in example[spans_key]:
                    if is_end_included:
                        span['end'] += 1
                    entities.append(span)

                mask = [True] * len(tokens)

                reading_format = corpus_config.get('encoding_format', 'concat')
                if reading_format == 'concat':
                    for entity in entities:
                        if cand is None:
                            entity['candidates'] = None
                        else:
                            if type(cand[guid]) == np.ndarray:
                                candidates = cand[guid].tolist()
                            else:
                                candidates = cand[guid]
                            entity['candidates'] = candidates
                        yield guid, {
                            'id': str(guid),
                            'tokens': tokens,
                            'spans': [entity],
                            'mask': mask,
                        }
                        guid += 1
                else:
                    raise NotImplementedError('unimplemented reading_format')



import os

from modelscope.hub.check_model import check_local_model_is_latest
from modelscope.hub.snapshot_download import snapshot_download
from modelscope.utils.constant import Invoke


def get_or_download_model_dir(model, model_revision=None):
    """get model cache dir"""
    if os.path.exists(model):
        model_cache_dir = model if os.path.isdir(model) else os.path.dirname(model)
        check_local_model_is_latest(model_cache_dir, user_agent={Invoke.KEY: Invoke.LOCAL_TRAINER})
    else:
        model_cache_dir = snapshot_download(
            model, revision=model_revision, user_agent={Invoke.KEY: Invoke.TRAINER}
        )
    return model_cache_dir



# Copyright (c) Alibaba, Inc. and its affiliates.
from modelscope.preprocessors.nlp.transformers_tokenizer import NLPTokenizer
from transformers import AutoTokenizer, BertTokenizer, BertTokenizerFast
from transformers.tokenization_utils import PreTrainedTokenizer

from adaseq.utils.hub_utils import get_or_download_model_dir


def build_tokenizer(
    model_name_or_path: str, use_fast: bool = True, is_word2vec: bool = False, **kwargs
) -> PreTrainedTokenizer:
    """build tokenizer from `transformers`."""

    if is_word2vec:
        tokenizer = BertTokenizerFast if use_fast else BertTokenizer
        return tokenizer.from_pretrained(model_name_or_path, **kwargs)

    try:
        return AutoTokenizer.from_pretrained(model_name_or_path, use_fast=use_fast, **kwargs)

    except OSError:
        model_name_or_path = get_or_download_model_dir(model_name_or_path)
        return NLPTokenizer(model_name_or_path, use_fast=use_fast).tokenizer



# Copyright (c) Alibaba, Inc. and its affiliates.
from .default_trainer import DefaultTrainer, build_trainer
from .hooks.text_logger_hook import AdaSeqTextLoggerHook
from .typing_trainer import TyingTrainer



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Dict, List, Union

from modelscope.preprocessors.builder import PREPROCESSORS
from modelscope.utils.constant import Fields

from adaseq.metainfo import Preprocessors

from .nlp_preprocessor import NLPPreprocessor


@PREPROCESSORS.register_module(Fields.nlp, module_name=Preprocessors.twostage_preprocessor)
class TwoStagePreprocessor(NLPPreprocessor):
    """Preprocessor for twostage-ner.
    span targets are processed into mention_boundary, type_ids, ident_ids.
    examples:
        span: {'start':1, 'end':2, 'type': ['PER']}
        processed: {'mention_boundary': [[1], [2]], 'type_ids':[1], 'ident_ids': ['S-SPAN']]}
    """

    def __init__(self, model_dir: str, **kwargs) -> None:

        super().__init__(model_dir, return_offsets=True, **kwargs)

        label_to_id = kwargs.pop('label_to_id', None)
        labels = kwargs.pop('labels', None)
        if label_to_id is not None:
            self.typing_label_to_id = label_to_id
        elif labels is not None:
            self.typing_label_to_id = self.make_label_to_id(labels)
        else:
            raise ValueError('Must have one of `labels` or `label_to_id`')
        self.typing_id_to_label = {
            v: k for k, v in sorted(self.typing_label_to_id.items(), key=lambda x: x[1])
        }
        # make sure they are aligned.
        assert len(self.typing_id_to_label) not in self.typing_id_to_label
        assert len(self.typing_id_to_label) - 1 in self.typing_id_to_label

        self.ident_label_to_id = {'O': 0, 'B-SPAN': 1, 'I-SPAN': 2, 'E-SPAN': 3, 'S-SPAN': 4}
        self.ident_id_to_label = {0: 'O', 1: 'B-SPAN', 2: 'I-SPAN', 3: 'E-SPAN', 4: 'S-SPAN'}

    def __call__(self, data: Union[str, List, Dict]) -> Dict[str, Any]:
        """prepare inputs for two-stage-ner model"""
        output = super().__call__(data)
        if isinstance(data, Dict):
            # span detection inputs
            output['ident_ids'] = self.__span2bioes(
                len(output['tokens']['mask']) - 2, data['spans']
            )
            # entity typing inputs
            self._generate_mentions_labels(data, output)
        return output

    def _generate_mentions_labels(self, data: Union[str, List, Dict], output: Dict[str, Any]):
        mention_type_ids = []
        boundary_starts = []
        boundary_ends = []
        mention_mask = []
        for span in data['spans']:
            if (
                span['start'] > len(output['tokens']['mask']) - 2
                or span['end'] > len(output['tokens']['mask']) - 2
            ):
                continue
            boundary_starts.append(span['start'])
            boundary_ends.append(span['end'] - 1)
            mention_mask.append(1)
            mention_type_ids.append(self.typing_label_to_id.get(span['type'], -1))

        output['mention_boundary'] = [boundary_starts, boundary_ends]
        output['mention_mask'] = mention_mask
        output['type_ids'] = mention_type_ids

    def __span2bioes(self, sequence_length: int, spans: List[Dict]) -> List[int]:
        ident_ids = [self.ident_label_to_id['O']] * sequence_length
        for span in spans:
            start = span['start']
            end = span['end']
            if start > sequence_length or end > sequence_length:
                continue
            if end - start == 1:
                ident_ids[start] = self.ident_label_to_id['S-SPAN']
            else:
                for i in range(start, end):
                    ident_ids[i] = self.ident_label_to_id['I-SPAN']
                ident_ids[start] = self.ident_label_to_id['B-SPAN']
                ident_ids[end - 1] = self.ident_label_to_id['E-SPAN']
        return ident_ids



# Copyright (c) Alibaba, Inc. and its affiliates.

from typing import Any, Dict

from modelscope.outputs import OutputKeys
from modelscope.pipelines.builder import PIPELINES
from modelscope.utils.tensor_utils import torch_nested_detach, torch_nested_numpify

from adaseq.metainfo import Pipelines, Tasks

from .base import Pipeline


@PIPELINES.register_module(
    Tasks.named_entity_recognition, module_name=Pipelines.sequence_labeling_pipeline
)
class SequenceLabelingPipeline(Pipeline):
    """
    A pipeline for sequence labeling models
    Model outputs should contain:
        predicts: [[0, 0, 1, 2, 2, ...]]
    Pipeline outputs will be like:
        {"output": [{"span": "国正", "type": "PER", "start": 0, "end": 2}, ...]}
    """

    def postprocess(  # noqa: D102
        self, inputs: Dict[str, Any], **postprocess_params
    ) -> Dict[str, Any]:
        text = inputs['meta']['text']

        # TODO post_process does not support batch for now.
        predictions = inputs['predicts']  # predicts: [[0, 0, 1, 2, 2, ...]]
        if len(predictions.shape) == 2:
            predictions = predictions[0]
        predictions = torch_nested_numpify(torch_nested_detach(predictions))

        offset_mapping = inputs['tokens']['offset_mapping']
        if offset_mapping.dim() == 3:
            offset_mapping = offset_mapping[0]
        offset_mapping = torch_nested_numpify(torch_nested_detach(offset_mapping))

        labels = [self.id2label[x] for x in predictions]

        outputs = []
        chunk = {}
        for label, offsets in zip(labels, offset_mapping):
            if label[0] in 'BS':
                if chunk:
                    chunk['span'] = text[chunk['start'] : chunk['end']]
                    outputs.append(chunk)
                chunk = {}

            if label[0] in 'BIES':
                if not chunk:
                    chunk = {'type': label[2:], 'start': offsets[0], 'end': offsets[1]}

            if label[0] in 'IES':
                if chunk:
                    chunk['end'] = offsets[1]

            if label[0] in 'ES':
                if chunk:
                    chunk['span'] = text[chunk['start'] : chunk['end']]
                    outputs.append(chunk)
                    chunk = {}

        if chunk:
            chunk['span'] = text[chunk['start'] : chunk['end']]
            outputs.append(chunk)

        return {OutputKeys.OUTPUT: outputs}



# Copyright (c) Alibaba, Inc. and its affiliates.
# Fix some modelscope bugs temporarily.
# These bugs will be fixed in the next modelscope version.


def suppress_modelscope_ast_warning():  # noqa
    try:
        from modelscope.utils.logger import get_logger

        def filter_modelscope_ast_warning(record):
            return 'not found in ast index file' not in record.msg

        logger = get_logger()
        logger.addFilter(filter_modelscope_ast_warning)
    except IsADirectoryError:
        pass


suppress_modelscope_ast_warning()



# Copyright (c) Alibaba, Inc. and its affiliates.
import json
import logging

from modelscope.preprocessors.base import Preprocessor
from modelscope.trainers.builder import TRAINERS

from adaseq.data.data_collators.base import DataCollatorWithPadding
from adaseq.data.dataset_manager import DatasetManager
from adaseq.metainfo import Trainers

from .default_trainer import DefaultTrainer

logger = logging.getLogger(__name__)


@TRAINERS.register_module(module_name=Trainers.typing_trainer)
class TyingTrainer(DefaultTrainer):
    """Default trainer class for AdaSeq.

    This trainer inherits from EpochBasedTrainer with some modifications.
    It implements some common data processing functions which are convenient for training a model.
    It also implements a basic test function for evaluate a trained model on the test dataset.

    Args:

    cfg_file (`str`): required
        The path of `Config` of this trial.
    work_dir (`str`): required
        The created directionary to save all produced files in training.
    dataset_manager (`DatasetManager`): required
        A `DatasetManager` with already loaded datasets.
    data_collator (`DataCollatorWithPadding`): required
        A `DataCollatorWithPadding` object to collect data batches.
    preprocessor (`Preprocessor`): required
        A `Preprocessor` to process data before training.
    seed (`int`): required
        Random seed.
    device (`str`): 'cpu' by default
        A device string, e.g., 'cpu', 'cuda', 'cuda:1'.
    kwargs (`dict`):
        Other keyword arguments for `EpochBasedTrainer` of `modelscope`.
    """

    def __init__(
        self,
        cfg_file: str,
        work_dir: str,
        dataset_manager: DatasetManager,
        data_collator: DataCollatorWithPadding,
        preprocessor: Preprocessor,
        seed: int = 42,
        device: str = 'cpu',
        **kwargs,
    ) -> None:
        super().__init__(
            cfg_file=cfg_file,
            work_dir=work_dir,
            dataset_manager=dataset_manager,
            data_collator=data_collator,
            preprocessor=preprocessor,
            seed=seed,
            device=device,
            **kwargs,
        )
        self.model.trainer = self  # add trainer to the model for config
        self.model.expand_vocab()

    def test(self, checkpoint_path=None):
        """Evaluate a trained model on testing set"""
        backup_eval_dataset = self.eval_dataset
        # first evaluate then test
        metric_values = self.evaluate(checkpoint_path)
        # log to terminal
        metric_string = json.dumps(metric_values, indent=2, ensure_ascii=False)
        logger.info('dev: ' + metric_string)

        self.do_test = True
        self.eval_dataset = self.test_dataset
        metric_values = self.evaluate(checkpoint_path)
        self.do_test = False
        self.eval_dataset = backup_eval_dataset

        # log to terminal
        metric_string = json.dumps(metric_values, indent=2, ensure_ascii=False)
        logger.info('test: ' + metric_string)

        # log to file
        self._init_file_logger()
        from collections import OrderedDict

        log_dict = OrderedDict(mode='test', seed=self._seed, **metric_values)
        self.dump_log(log_dict)

        return metric_values



# Copyright (c) Alibaba, Inc. and its affiliates.
import logging
from typing import Dict

import numpy as np
import torch
from modelscope.metrics.base import Metric
from modelscope.metrics.builder import METRICS, MetricKeys

from adaseq.metainfo import Metrics

logger = logging.getLogger(__name__)


class SetScore:
    """evaluate macro and micro set p/r/f1 scores"""

    def __init__(self):
        self.n_sample = 0
        self.pred = []  # list of list
        self.true = []  # list of list

    def reset(self):  # noqa: D102
        self.n_sample = 0
        self.pred = []  # list of list
        self.true = []  # list of list

    def set_pred_true(self, pred, true):  # noqa: D102
        self.pred = pred
        self.true = true

    def update(self, batch_gold_entities, batch_pred_entities):  # noqa: D102
        self.n_sample += len(batch_gold_entities)
        self.pred.extend(batch_pred_entities)
        self.true.extend(batch_gold_entities)

    def f1(self, precision, recall):  # noqa: D102
        f1 = 0.0 if recall + precision == 0 else (2 * precision * recall) / (precision + recall)
        return f1

    def result(self):  # noqa: D102
        assert len(self.pred) == len(self.true)
        M = len(self.pred)
        strict_acc = 0
        num_pred_labels = 0
        num_true_labels = 0
        num_correct_labels = 0
        total_ma_p = 0
        total_ma_r = 0
        total_ma_f1 = 0
        count = 0
        for i in range(M):
            p = set(self.pred[i])
            t = set(self.true[i])
            count += 1

            if p == t:
                strict_acc += 1

            l_p, l_t, l_intersect = len(p), len(t), len(p.intersection(t))
            num_pred_labels += l_p
            num_true_labels += l_t
            num_correct_labels += l_intersect

            if l_p == 0 or l_t == 0:
                ma_p = 0
                ma_r = 0
                ma_f1 = 0
            else:
                ma_p = l_intersect / l_p
                ma_r = l_intersect / l_t
                ma_f1 = self.f1(ma_p, ma_r)

            total_ma_p += ma_p
            total_ma_r += ma_r
            total_ma_f1 += ma_f1

        if num_pred_labels == 0 or num_true_labels == 0:
            micro_p = 0
            micro_r = 0
            micro_f1 = 0
        else:
            micro_p = num_correct_labels / num_pred_labels
            micro_r = num_correct_labels / num_true_labels
            micro_f1 = self.f1(micro_p, micro_r)

        strict_acc /= count
        macro_p = total_ma_p / count
        macro_r = total_ma_r / count
        macro_f1 = self.f1(macro_p, macro_r)
        avg_true_label = num_true_labels / M
        avg_pred_label = num_pred_labels / M

        return {
            'strict_acc': strict_acc,
            'micro_p': micro_p,
            'micro_r': micro_r,
            'micro_f1': micro_f1,
            'macro_p': macro_p,
            'macro_r': macro_r,
            'macro_f1': macro_f1,
            'avg_true_label': avg_true_label,
            'avg_pred_label': avg_pred_label,
        }


@METRICS.register_module(module_name=Metrics.typing_metric)
class TypingMetric(Metric):
    """Evaluate metrics for typing tasks"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.scorer = SetScore()

    def add(self, outputs: Dict, inputs: Dict):  # noqa: D102
        predicts = outputs['predicts']
        pred_results = list()
        ground_truths = list()
        for i, meta in enumerate(inputs['meta']):
            for j, s in enumerate(meta['spans']):
                pred_results.append(predicts[i][j])
                ground_truths.append(set(s['type']))
        self.scorer.update(ground_truths, pred_results)

    def evaluate(self):  # noqa: D102
        score_detail = self.scorer.result()
        scores = {}
        scores[MetricKeys.PRECISION] = score_detail['macro_p']
        scores[MetricKeys.RECALL] = score_detail['macro_r']
        scores[MetricKeys.F1] = score_detail['macro_f1']
        scores[MetricKeys.ACCURACY] = score_detail['strict_acc']
        scores['micro_f1'] = score_detail['micro_f1']
        scores['avg_true_label'] = score_detail['avg_true_label']
        scores['avg_pred_label'] = score_detail['avg_pred_label']

        return scores

    def merge(self, other):  # noqa: D102
        raise NotImplementedError


@METRICS.register_module(module_name=Metrics.typing_threshold_metric)
class ConcatTypingThresholdMetric(Metric):
    """Evaluate metrics for typing tasks, with threshold tuning"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.scorer = SetScore()
        self.all_logits = []
        self.all_cands = []
        self.all_true = []

    def add(self, outputs: Dict, inputs: Dict):  # noqa: D102
        batch_logits = outputs['logits'].sigmoid().detach().cpu().numpy()
        B, C = batch_logits.shape[0], batch_logits.shape[-1]
        label_to_id = outputs['label_to_id']  # convert labels to id for easier evaluation

        if 'cands' in inputs:
            batch_cands = inputs['cands'].cpu().numpy()
        else:
            batch_cands = torch.arange(C).repeat(B, 1).cpu().numpy()

        ground_truths = list()
        for i, meta in enumerate(inputs['meta']):
            for j, s in enumerate(meta['spans']):
                ground_truths.append(set([label_to_id[j] for j in s['type']]))

        self.all_true += ground_truths
        self.all_cands.append(batch_cands)
        self.all_logits.append(batch_logits)

    def tune_threshold(self):
        """
        TODO
        """
        th = np.linspace(self.all_logits.min(), self.all_logits.max(), 10)
        all_res = []

        all_res_detail = []
        # stage 1 tuning
        for _th in th:
            pred_ = [
                set(self.all_cands[i][self.all_logits[i] > _th])
                for i in range(len(self.all_logits))
            ]
            self.scorer.set_pred_true(pred_, self.all_true)
            res = self.scorer.result()
            res_val = res['macro_f1'].__round__(5)
            all_res.append(res_val)
            all_res_detail.append(res)

        logger.info('phase 1: all thresholds: {}'.format(th))
        logger.info('phase 1: all results: {}'.format(all_res))

        best = np.argsort(all_res)[-1]
        best_1_th = th[best]
        new_range = (max(best_1_th - 0.15, 0), min(best_1_th + 0.15, 1.0))
        step = (new_range[1] - new_range[0]) / 30
        new_th = np.arange(new_range[0], new_range[1], step)
        all_res = []
        all_res_detail = []

        for _th in new_th:
            pred_ = [
                set(self.all_cands[i][self.all_logits[i] > _th])
                for i in range(len(self.all_logits))
            ]
            self.scorer.set_pred_true(pred_, self.all_true)
            res = self.scorer.result()
            res_val = res['macro_f1'].__round__(5)
            all_res.append(res_val)
            all_res_detail.append(res)

        best_th, best_th_res = new_th[np.argmax(all_res)], all_res_detail[np.argmax(all_res)]
        logger.info('phase 2: all thresholds: {}'.format(new_th))
        logger.info('phase 2: all results: {}'.format(all_res))

        logger.info('phase 2: best threshold: {}'.format(best_th))

        return best_th, best_th_res

    def apply_threshold(self, th):
        """
        TODO
        """
        pred_ = [
            set(self.all_cands[i][self.all_logits[i] > th]) for i in range(len(self.all_logits))
        ]
        self.scorer.set_pred_true(pred_, self.all_true)
        res = self.scorer.result()
        return th, res

    def evaluate(self):  # noqa: D102
        self.all_cands = np.concatenate(self.all_cands)
        self.all_logits = np.concatenate(self.all_logits)

        if getattr(self.trainer, 'do_test', False):
            best_th, score_detail = self.apply_threshold(th=getattr(self.trainer, 'best_th', 0.5))
        else:
            best_th, score_detail = self.tune_threshold()
            self.trainer.best_th = best_th

        scores = {}
        scores[MetricKeys.PRECISION] = score_detail['macro_p']
        scores[MetricKeys.RECALL] = score_detail['macro_r']
        scores[MetricKeys.F1] = score_detail['macro_f1']
        scores[MetricKeys.ACCURACY] = score_detail['strict_acc']
        scores['micro_f1'] = score_detail['micro_f1']
        scores['avg_true_label'] = score_detail['avg_true_label']
        scores['avg_pred_label'] = score_detail['avg_pred_label']

        return scores

    def merge(self, other):  # noqa: D102
        raise NotImplementedError
